QQ:=Rationals();
ZZ:=Integers();
P<q>:=PuiseuxSeriesRing(QQ);



function Eta(m : Prec:=100)
	eta:=1;
	B:=Ceiling(1/6+Sqrt(2/3*Prec+1/36))+1;
	for k in [1..B] do
		eta+:=(-1)^k*(q^(m*(k*(3*k+1)) div 2)+q^(m*(k*(3*k-1)) div 2));
	end for;
	return q^(m/24)*eta+O(q^(m*Prec));
end function;


/*function Collapse(f)
	c,s:=Coefficients(f);
	M:=AbsolutePrecision(f);
	L:=#c;
	supp:=[n+s-1 : n in [1..L] | c[n] ne 0];
	d:=GCD(supp);
	g:=O(q^(M div d));
	for n in [1..L] do
		g+:=c[n]*q^((n+s-1) div d);
	end for;
	return g,d;
end function;*/

function EtaQuotient(etalist : Prec:=100) 
	f:=1;
	gamma:=Matrix(2,[1,0,0,1]);
	L:=[l[2]*GCD(l[1],gamma[2][1])^2/l[1] : l in etalist | l[2] lt 0];
	if L eq [] then
		a:=0;
	else
		a:=Floor(1/24*&+L)-1;
	end if;
	
	for l in etalist do
		if l[2] ne 0 then
			f*:=(Eta(l[1] : Prec:=Ceiling((Prec-a+1)/l[1]))+O(q^(Prec-a+1)))^l[2];
		end if;
	end for;
	//print f;
	//g:=Collapse(f,24);
	if AbsolutePrecision(f) lt 0 then
		error "Precision is not high enough.";
	end if;	
	return f;
end function;


function P1(x)
	if IsIntegral(x) then
		return 0;
	else
		return FractionalPart(x)-1/2;
	end if;
end function;


function DedekindSum(h,k)
	s:=0;
	for m in [1..k] do
		s+:=P1(h*m/k)*P1(m/k);
	end for;
	return s;
end function;


function Rademacher(g) // g=matrix([a,b;c,d]) in SL2(ZZ)
	a:=g[1][1];
	b:=g[1][2];
	c:=g[2][1];
	d:=g[2][2];
	if c eq 0 then
		return b/d;
	else 	
		return (a+d)/c-12*Sign(c)*DedekindSum(d,Abs(c));
	end if;
end function;

function SymbolUp(c,d)
	if not (GCD(c,d) eq 1 and IsOdd(d)) then
		error "Invalid arguments";
	elif c ne 0 then
		return KroneckerSymbol(c,Abs(d));
	else
		return 1;
	end if;
end function;

function SymbolDown(c,d)
	if not (GCD(c,d) eq 1 and IsOdd(d)) then
		error "Invalid arguments";
	elif c ne 0 then
		return KroneckerSymbol(c,Abs(d))*(-1)^(((Sign(c)-1)*(Sign(d)-1)) div 4);
	else
		return d;
	end if;
end function;

function EtaMultiplier(g)	
	a:=g[1][1];
	b:=g[1][2];
	c:=g[2][1];
	d:=g[2][2];
	zeta:=RootOfUnity(24);
	if IsOdd(c) then
		return SymbolUp(d,c)*zeta^((a+d)*c-b*d*(c^2-1)-3*c);
	elif IsEven(c) then
		return SymbolDown(c,d)*zeta^((a+d)*c-b*d*(c^2-1)+3*d-3-3*c*d);
	end if;
end function;
	


function EtaProductSymbol(g,etalist)     // etalist=[<d,r> : d divides Level, r exponents], <d,r> corresponds to eta(d\tau)^r.
	N:=LCM([l[1] : l in etalist]);   // the level of the eta product
	k:=&+[l[2] : l in etalist];      // the weight of the eta product is k/2
	print k;
	a:=g[1][1];
	b:=g[1][2];
	c:=g[2][1];
	d:=g[2][2];
	if c mod N ne 0 then
		printf "The matrix must be in Gamma0(%o).\n",N;
		return "Error";
	end if;
	s:=0;
	for l in etalist do
		gm:=g;
		gm[1][2]:=l[1]*b;
		gm[2][1]:=c div l[1];
		s+:=Rademacher(gm);
	end for;
	return s*k/24;
end function;

printf "Enter the level N of the eta quotients in question:";
readi N;


K:=CyclotomicField(24*N);
R<qq>:=PuiseuxSeriesRing(K); 

function MySqrt(n) // Compute the sqrt of an integer exactly in terms of roots of unity
	F:=Factorization(n);
	s:=1;
	for f in F do
		p:=f[1];
		e:=f[2];
		s*:=p^(e div 2);
		if p eq 2 and IsOdd(e) then
			s*:=(1+RootOfUnity(4))/RootOfUnity(8);
		elif IsOdd(p) and IsOdd(e) then
			z:=RootOfUnity(p);
			r:=0;
			for k in [1..p-1] do
				l:=k^2 mod p;
				r+:=z^l;
			end for;
			if p mod 4 eq 1 then
				r+:=1;
			else
				r:=(r+1)/RootOfUnity(4);
			end if;
			s*:=r;
		end if;
	end for;
	assert s^2 eq n;
	return s;
end function;


function EtaAtCusp(m,gamma : Prec:=20) // Compute the Fourier expansion of eta(m\tau)|gamma.
	a:=gamma[1][1];
	b:=gamma[1][2];
	c:=gamma[2][1];
	d:=gamma[2][2];
	g:=GCD(c,m);
	x:=m*a div g;
	u:=c div g;
	_,y,v:=XGCD(u,x);
	y:=-y;
	nu:=m*b*v-y*d;
	L:=Matrix(2,[x,y,u,v]);
	v:=EtaMultiplier(L);
	eta:=0;
	Prec:=Max([Prec,g^2/(24*m)+1]);
	B:=Ceiling(Sqrt(Prec*24*m)/g);
	//print B;
	for n in [1..B] do
		if GCD(12,n) eq 1 then
			eta+:=KroneckerSymbol(12,n)*qq^((n*g)^2/(24*m))*K!RootOfUnity(24*m)^ZZ!(n^2*nu*g);
		end if;
	end for;
	r:=Numerator(g/m);
	s:=Denominator(g/m);
	f:=v*MySqrt(r)/MySqrt(s)*eta+O(qq^Prec);
	if IsWeaklyZero(f) then
		error "Precision not high enough.";
	end if;
	return f;
end function;

function EtaQuotientAtCusp(etalist, gamma : Prec:=20, w:=1) // w stands for the width of the cusp
	f:=1;
	L:=[l[2]*GCD(l[1],gamma[2][1])^2/l[1] : l in etalist | l[2] lt 0];
	if L eq [] then
		a:=0;
	else
		a:=Floor(1/24*&+L)-1;
	end if;
	for l in etalist do
		if l[2] ne 0 then
			f*:=(EtaAtCusp(l[1],gamma : Prec:=(Prec+1)/w-a)+O(qq^((Prec+1)/w-a)))^l[2];
		end if;
	end for;
	//print f;
	//g:=Collapse(f,24);
	if AbsolutePrecision(f) lt 0 then
		error "Precision is not high enough.";
	end if;
	return f;
end function;


function MoonshineMultiplier(gamma,v,h,N)
	if not gamma in Gamma0(N) then
		error "Invalid arguments";
	end if;
	c:=gamma[2][1];
	d:=gamma[2][2];
	return RootOfUnity(h)^(-v*c*d div N);
end function;


function VOperator(f,d)
	c,s,w:=Coefficients(f);
	Prec:=AbsolutePrecision(f);
	g:=0;
	for j in [1..#c] do
		g+:=c[j]*qq^((j+s-1)/w*d);
	end for;
	return g+O(qq^(Prec*d));
end function;

load 'EtaWt2/EtaWt2_108.m';
load 'EtaWt2/EtaWt2_128.m';
load 'EtaWt2/EtaWt2_288.m';
load 'EtaWt2/EtaWt2_180.m';
load 'EtaWt2/EtaWt2_216.m';
load 'EtaWt2/EtaWt2_252.m';
load 'EtaWt2/EtaWt2_576.m';
load 'EtaWt2/EtaWt2_1152.m';
load 'EtaWt2/EtaWt2_324.m';
load 'EtaWt2/EtaWt2_360.m';
load 'EtaWt2/EtaWt2_468.m';

function FindFormsWithMultiplier(basis,level,N,v,h : Prec:=12)
	C:=[Matrix(g) : g in CosetRepresentatives(Gamma0(level)) | g in Gamma0(N) and not g in Gamma0(level)];
	d:=#basis;
	V:=VectorSpace(K,d);
	W:=V;
	for gamma in C do
		printf "Beginning to check class %o out of %o.\n", Index(C,gamma),#C;
		glist:=[];
		wlist:=[];
		min:=Minimum([l[2] : l in f, f in basis] cat [0]);
		qexps:=[EtaAtCusp(m,gamma : Prec:=d+4-min) : m in Divisors(level)];
		printf "Done building q-expansions of eta.\n";
		// Try to work in  a smaller field to speed up arithmetic
		L:=[Minimise(Coefficients(f)) : f in qexps];
		F:=Universe(Random(L));
		printf "We can work in a cyclotomic subfield of K of index %o.\n",Degree(K)/Degree(F);
		P<q>:=PuiseuxSeriesRing(F);
		qexps:=[P!f : f in qexps];
		for f in basis do
			g:=&*[qexps[j]^f[j][2] : j in [1..#f]]-MoonshineMultiplier(gamma,v,h,N)*R!EtaQuotient(f : Prec:=d+4);
			printf "Done building eta quotient %o out of %o.\n",Index(basis,f),d;
			if AbsolutePrecision(g) le d then
				error "Precision not high enough.";
			end if;
			_,_,w:=Coefficients(g);
			Append(~glist,g);
			Append(~wlist,w);
		end for;
		width:=LCM(wlist);
		A:=[];
		for j in [1..d] do
			g:=P!VOperator(glist[j],width);
			Prec:=AbsolutePrecision(g);
			c,s:=Coefficients(g);
			c:=[0^^s] cat c;
			c cat:=[0^^(Prec-#c)];
			S:=c[1..width*d];
			Append(~A,S);
		end for;
		A:=Matrix(A);
		try 
			V:=ChangeRing(V,F);
			W:=sub<V | ChangeRing(W,F)>;
			W meet:=sub<V | NullSpace(A)>;
		catch e
			printf "Not possible to work in a smaller field.\n";
			V:=ChangeRing(V,K);
			W:=ChangeRing(W,K);	
			W meet:=sub<V | ChangeRing(NullSpace(A),K)>;
		end try;
		printf "Done checking class %o out of %o. The space of of possible forms has dimension %o.\n",Index(C,gamma),#C,Dimension(W);
		if Dimension(W) eq 0 then	
			return [],[];
		end if;
	end for;
	/*B:=[];
	for v in Basis(W) do
		g:=&+[v[j]*R!EtaQuotient(basis[j] : Prec:=Prec) : j in [1..d]];
		Append(~B,g);
	end for;*/
	return Basis(W);
end function;
			
function DivideBasis(Basis,g,N) // Divide all eta quotients in Basis by eta quotient g
	NewBasis:=[];
	L:=[a[1] : a in g];
	for k in [1..#Basis] do
		b:=Basis[k];
		for i in [1..#b] do
			j:=Index(L,b[i][1]);
			if j ne 0 then
				b[i][2]-:=g[j][2];
			end if;
		end for;	
		Append(~NewBasis,b);
	end for;
	/*filename:="EtaWtHalf.sage";
	name:="M_" cat IntegerToString(N) cat "=";
	PrintFile(filename,name);
	PrintFile(filename,NewBasis);
	PrintFile(filename,"\n");*/
	return NewBasis;
end function;


function FindCuspFormsWithMultiplier(basis,level,N,v,h : Prec:=20, Plus:=true) // Compute a basis of weight 3/2 cusp forms with the given Moonshine multiplier.
	EtaCoeffs:=FindFormsWithMultiplier(basis,level,N,v,h : Prec:=Prec);
	printf "The space of modular forms of weight 2 on Gamma0(%o) with the given multiplier has dimension %o.\n",N,#EtaCoeffs;
	if #EtaCoeffs eq 0 then
		return [],[],[],[];
	end if;
	d:=#basis;
	V:=VectorSpace(K,#EtaCoeffs);
	W:=V;
	basis32:=DivideBasis(basis,[<1,-2>,<2,5>,<4,-2>],N);
	HH:=UpperHalfPlaneWithCusps();
	C:=Cusps(Gamma0(N));
	for c in C do
		printf "Examining cusp %o.\n",c;
		_, gamma:=EquivalentPoint(HH!c);
		gamma:=Matrix(gamma);
		w:=CuspWidth(Gamma0(N),c);
		fList:=[];
		min:=Minimum([l[2] : l in f, f in basis] cat [0]);
		qexps:=[EtaAtCusp(m,gamma^-1 : Prec:=(2-min)/w) : m in Divisors(level)];
		printf "Done building q-expansions of eta.\n";
		// Try to work in  a smaller field to speed up arithmetic
		L:=[Minimise(Coefficients(f)) : f in qexps];
		F:=Universe(Random(L));
		printf "We can work in a cyclotomic subfield of K of index %o.\n",Degree(K)/Degree(F);
		P<q>:=PuiseuxSeriesRing(F);
		qexps:=[P!f : f in qexps];
		for v in EtaCoeffs do
			f:=0;
			for j in [1..d] do
				if v[j] ne 0 then
					g:=&*[qexps[i]^basis32[j][i][2] : i in [1..#qexps]];
					// printf "Computed eta quotient %o out of %o.\n",j,d;
					if AbsolutePrecision(g) le 0 then
						error "Precision not high enough.";
					end if;
					f+:=F!v[j]*P!VOperator(g,w)+O(q);
				end if;
			end for;
			Append(~fList,f);
		end for;
		min:=Minimum([Valuation(f) : f in fList]);
		m:=Denominator(min);
		if m gt 1 then
			printf "The cusp %o is exceptional.\n",c;
			// print fList;
			fList:=[P!VOperator(f,m) : f in fList];
			min:=(min*m);
		end if;
		min:=ZZ!min;
		if min gt 0 then
			U:=V;
		else
			A:=[];
			for f in fList do
				if IsWeaklyZero(f) then
					Append(~A,[K!0^^(1-min)]);
				else
					coeffs,s:=Coefficients(qq^(-min)*f);
					coeffs:=[0^^s] cat coeffs;
					coeffs cat:=[0^^(Ceiling(AbsolutePrecision(qq^(-min)*f))-#coeffs)];
					S:=Minimise(coeffs[1..-min+1]);
					Append(~A,S);
				end if;
			end for;
			A:=Matrix(A);
			try 
				V:=ChangeRing(V,F);
				W:=ChangeRing(W,F);
				U:=sub<V | NullSpace(A)>;
			catch e
				printf "Not possible to work in a smaller field.\n";
				V:=ChangeRing(V,K);
				W:=ChangeRing(W,K);
				U:=sub<V | ChangeRing(NullSpace(A),K)>;
			end try;
		end if;
		W:=W meet U;
		printf "Done checking cusp %o out of %o. The space of of possible forms has dimension %o.\n",Index(C,c),#C,Dimension(W);
		if Dimension(W) eq 0 then	
			return [],[],[],[];
		end if;
	end for;
	BM:=ChangeRing(BasisMatrix(W),QQ);
	A:=ChangeRing(Matrix(EtaCoeffs),QQ);
	//print BM,A;
	NewBasis:=Rows(BM*A);
	B:=[];
	for v in NewBasis do
		g:=&+[QQ!v[j]*EtaQuotient(basis32[j] : Prec:=Prec) : j in [1..d]];
		prec:=AbsolutePrecision(g);
		if not IsIntegral(prec) then
			prec:=Ceiling(prec);
		end if;
		g:=Truncate(g);
		Append(~B,g+O(qq^prec));
	end for;
	PlusBasis:=[];
	PB:=[];
	if Plus then
		print "Starting computation of plus space basis.";
		A12:=[];
		for f in B do
			coeffs,s:=Coefficients(f);
			coeffs:=[0^^(s-1)] cat coeffs;
			coeffs cat:=[0^^(AbsolutePrecision(f)-#coeffs-1)];
			row:=[coeffs[j] : j in [1..Prec] | j mod 4 in [1,2]];
			Append(~A12,row);
		end for;
		A12:=ChangeRing(Matrix(A12),QQ);
		CM:=BasisMatrix(NullSpace(A12));
		PlusBasis:=Rows(CM*BM*A);
		for v in PlusBasis do
			g:=&+[QQ!v[j]*EtaQuotient(basis32[j] : Prec:=Prec) : j in [1..d]];
			prec:=AbsolutePrecision(g);
			if not IsIntegral(prec) then
				prec:=Ceiling(prec);
			end if;
			g:=Truncate(g);
			Append(~PB,g+O(qq^prec));
		end for;
	end if;
	return B,NewBasis,PB,PlusBasis;
end function;


function FindFormsWithMultiplier12(basis,level,N,v,h : Prec:=20, Plus:=true) // Find a basis of weight 1/2 holomorphic modular forms with the given Moonshine multiplier.
	EtaCoeffs:=FindFormsWithMultiplier(basis,level,N,v,h : Prec:=Prec);
	printf "The space of modular forms of weight 2 on Gamma0(%o) with the given multiplier has dimension %o.\n",N,#EtaCoeffs;
	if #EtaCoeffs eq 0 then
		return [],[],[],[];
	end if;
	d:=#basis;
	V:=VectorSpace(K,#EtaCoeffs);
	W:=V;
	basis12:=DivideBasis(basis,[<1,-6>,<2,15>,<4,-6>],N);
	HH:=UpperHalfPlaneWithCusps();
	C:=Cusps(Gamma0(N));
	for c in C do
		printf "Examining cusp %o.\n",c;
		_, gamma:=EquivalentPoint(HH!c);
		gamma:=Matrix(gamma);
		w:=CuspWidth(Gamma0(N),c);
		fList:=[];
		min:=Minimum([l[2] : l in f, f in basis] cat [0]);
		qexps:=[EtaAtCusp(m,gamma^-1 : Prec:=(2-min)/w) : m in Divisors(level)];
		printf "Done building q-expansions of eta.\n";
		// Try to work in  a smaller field to speed up arithmetic
		L:=[Minimise(Coefficients(f)) : f in qexps];
		F:=Universe(Random(L));
		printf "We can work in a cyclotomic subfield of K of index %o.\n",Degree(K)/Degree(F);
		P<q>:=PuiseuxSeriesRing(F);
		qexps:=[P!f : f in qexps];
		for v in EtaCoeffs do
			f:=0;
			for j in [1..d] do
				if v[j] ne 0 then
					g:=&*[qexps[i]^basis12[j][i][2] : i in [1..#qexps]];
					// printf "Computed eta quotient %o out of %o.\n",j,d;
					if AbsolutePrecision(g) le 0 then
						error "Precision not high enough.";
					end if;
					f+:=F!v[j]*P!VOperator(g,w)+O(q);
				end if;
			end for;
			Append(~fList,f);
		end for;
		min:=Minimum([Valuation(f) : f in fList]);
		m:=Denominator(min);
		if m gt 1 then
			printf "The cusp %o is exceptional.\n",c;
			// print fList;
			fList:=[P!VOperator(f,m) : f in fList];
			min:=(min*m);
		end if;
		min:=ZZ!min;
		if min ge 0 then
			U:=V;
		else
			A:=[];
			for f in fList do
				if IsWeaklyZero(f) then
					Append(~A,[K!0^^(1-min)]);
				else
					coeffs,s:=Coefficients(qq^(-min)*f);
					coeffs:=[0^^s] cat coeffs;
					coeffs cat:=[0^^(Ceiling(AbsolutePrecision(qq^(-min)*f))-#coeffs)];
					S:=Minimise(coeffs[1..-min+1]);
					Append(~A,S);
				end if;
			end for;
			A:=Matrix(A);
			try 
				V:=ChangeRing(V,F);
				W:=ChangeRing(W,F);
				U:=sub<V | NullSpace(A)>;
			catch e
				printf "Not possible to work in a smaller field.\n";
				V:=ChangeRing(V,K);
				W:=ChangeRing(W,K);
				U:=sub<V | ChangeRing(NullSpace(A),K)>;
			end try;
		end if;
		W:=W meet U;
		printf "Done checking cusp %o out of %o. The space of of possible forms has dimension %o.\n",Index(C,c),#C,Dimension(W);
		if Dimension(W) eq 0 then	
			return [],[],[],[];
		end if;
	end for;
	BM:=ChangeRing(BasisMatrix(W),QQ);
	A:=ChangeRing(Matrix(EtaCoeffs),QQ);
	//print BM,A;
	NewBasis:=Rows(BM*A);
	B:=[];
	for v in NewBasis do
		g:=&+[QQ!v[j]*EtaQuotient(basis12[j] : Prec:=Prec) : j in [1..d]];
		prec:=AbsolutePrecision(g);
		if not IsIntegral(prec) then
			prec:=Ceiling(prec);
		end if;
		g:=Truncate(g);
		Append(~B,g+O(qq^prec));
	end for;
	return B,NewBasis;
end function;
			
			

function CuspOrders(etalist,coeffs : Prec:=1) 
	g:=[<1,-2>,<2,5>,<4,-2>];
	etalist:=DivideBasis(etalist,g,N);
	HH:=UpperHalfPlaneWithCusps();
	C:=Cusps(Gamma0(N));
//	Exclude(~C,C[1]); // Exclude the cusp infinity.
	
	printf "Starting computation of expansions at cusps.\n";
	ords:=[];	
	for k in C do
		printf "Considering cusp %o.\n",k;
		_, gamma:=EquivalentPoint(HH!k);
		w:=CuspWidth(Gamma0(N),k);
		fk:=0;
		for j in [1..#coeffs] do
			if coeffs[j] ne 0 then
				printf "Starting to compute eta quotient %o out of %o.\n",j,#coeffs; 
				fk+:=coeffs[j]*EtaQuotientAtCusp(etalist[j], Matrix(gamma)^-1 : Prec:=Prec, w:=w);
			end if;
		end for;
		print fk;
		printf "The order of the given function at cusp %o is %o.\n",k,Valuation(fk);
		Append(~ords,<k,QQ!Valuation(fk)>);
	end for;
	return ords;
end function;

		
		
